/* $LAAS: vectorization.h 2010/12/10 11:39:40 mallet $ */

/* 
 * Copyright (c) 2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Sun Aug 24 2008
 */

#ifndef H_STEREO_VECTORIZATION
#define H_STEREO_VECTORIZATION

/** branch prediction */
#ifdef __GNUC__
# if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 96)
#  define likely(x)   (__builtin_expect((x) != 0, 1))
#  define unlikely(x)   (__builtin_expect((x), 0))
# else
#  define likely(x)   (x)
#  define unlikely(x)   (x)
# endif
#else
#  define likely(x)   (x)
#  define unlikely(x)   (x)
#endif


#if SPIX_SSE && defined(__SSE2__) && defined(__GNUC__)

#include <xmmintrin.h>
#include <emmintrin.h>

/** A 128bits vector of 'type' elements */
# define spix_vector(type)	type __attribute__((vector_size (16)))

/** Length of vectors */
# define spix_vector_len(type)	(sizeof(spix_vector(type))/sizeof(type))

/** Alignment in bytes */
# define spix_align_size	(16)

/** Alignment declaration for variables that must be accessed with vectors */
# define spix_aligned		__attribute__((aligned (spix_align_size)))

/** Check if a pointer is aligned */
# define spix_isaligned(x)	(!((unsigned int)(x) % spix_align_size))

/** Load unaligned data from memory */
# define spix_ldu(type, x)	(spix_vector(type))_mm_loadu_ps((float *)(x))

/** Load aligned data from memory */
# define spix_lda(type, x)	*(spix_vector(type) *)(x)

/** Store data to unaligned memory */
# define spix_stru(x, type, v)	_mm_storeu_ps((float *)(x), (spix_vector(float))(v))

/** Store data to aligned memory */
# define spix_stra(x, type, v)	*(spix_vector(type) *)(x) = (v)

/** Convert ints to floats */
# define spix_cvtips(x)		_mm_cvtepi32_ps(*(__m128i *)(x))

/** Make a vector with all the same elements */
# define spix_1float(x)		_mm_set1_ps(x)

/** Make a vector with all the same elements */
# define spix_1short(x)		(spix_vector(short))_mm_set1_epi16(x)

/** Make a vector with all the same elements */
# define spix_1char(x)		(spix_vector(short))_mm_set1_epi8(x)

/** Get max and max index for floats */
# define spix_maxind_float(x, k, m, mi)	do {				\
    spix_vector(float) mask = _mm_cmpgt_ps((x), m);			\
    if (_mm_movemask_epi8((spix_vector(int))mask)) {			\
      (m) = _mm_max_ps((x), (m));					\
      (mi) = _mm_or_ps(_mm_and_ps(mask,(k)), _mm_andnot_ps(mask,(mi)));	\
    }									\
} while(0)

/** Get max and max index for shorts */
# define spix_maxind_short(x, k, m, mi)	do {				\
    spix_vector(short) mask = _mm_cmpgt_epi16((x), m);			\
    if (_mm_movemask_epi8((spix_vector(int))mask)) {			\
      (m) = _mm_max_epi16((x), (m));					\
      (mi) = _mm_or_si128(						\
	_mm_and_si128(mask,(k)), _mm_andnot_si128(mask,(mi)));		\
    }									\
} while(0)

#else

/** A single element vector of 'type' elements */
# define spix_vector(type)	type

/** Length of vectors */
# define spix_vector_len(type)	(1)

/** Alignment in bytes */
# define spix_align_size	(sizeof(void *))

/** Alignment declaration for variables that must be accessed with vectors */
# define spix_aligned

/** Check if a pointer is aligned */
# define spix_isaligned(x)	(1)

/** Load unaligned data from memory */
# define spix_ldu(type, x)	(*(x))

/** Load aligned data from memory */
# define spix_lda(type, x)	(*(x))

/** Store data to unaligned memory */
# define spix_stru(x, type, v)	*(x) = (v)

/** Store data to aligned memory */
# define spix_stra(x, type, v)	*(x) = (v)

/** Convert ints to floats */
# define spix_cvtips(x)		(*(x))

/** Make a vector with all the same elements */
# define spix_1float(x)		(x)

/** Make a vector with all the same elements */
# define spix_1short(x)		(x)

/** Make a vector with all the same elements */
# define spix_1char(x)		(x)

/** Get max and max index for floats */
# define spix_maxind_float(x, k, m, mi)	do {				\
    if ((x) > (m)) {							\
      (m) = (x);							\
      (mi) = (k);							\
    }									\
} while(0)

/** Get max and max index for shorts */
# define spix_maxind_short(x, k, m, mi)	spix_maxind_float(x, k, m, mi)

#endif /* SPIX_SSE && __SSE__ && __GNUC__ */

#endif /* H_STEREO_VECTORIZATION */
